'use strict';

var store = require('svelte/store');
var zod = require('zod');

function zip(...arrays) {
  if (!arrays || !arrays.length)
    return [];
  return new Array(Math.max(...arrays.map(({ length }) => length))).fill([]).map((_, idx) => arrays.map((array) => array[idx]));
}

const debounce = ({ delay }, func) => {
  let timer = void 0;
  let active = true;
  const debounced = (...args) => {
    if (active) {
      clearTimeout(timer);
      timer = setTimeout(() => {
        active && func(...args);
      }, delay);
    } else {
      func(...args);
    }
  };
  debounced.cancel = () => {
    active = false;
  };
  debounced.flush = (...args) => func(...args);
  return debounced;
};

const trurthly = (z) => !!z;
const toReadable = (w) => store.derived(w, a => a);
const getErrorMessage = (e) => {
    if (e instanceof zod.ZodError)
        return e.issues
            .filter(a => a.path.length === 0) // Only take error from root
            .map(a => a.message)
            .join();
    if (e instanceof Error)
        return e.message;
    if (typeof e === 'string')
        return e;
    if (typeof e === 'undefined')
        return '';
    if (e === null)
        return '';
    return JSON.stringify(e);
};

class ZodFieldStore {
    name;
    value;
    touched;
    dirty;
    error;
    valid;
    handleUpdate;
    handleChange;
    handleBlur;
    reset;
    setError;
    setTouched;
    constructor(parentSchema, name, initialValue, ms = 0) {
        if (!(name in parentSchema.shape))
            throw new Error('Invalid field name');
        // Stores
        const touched = store.writable(false);
        const dirty = store.writable(false);
        const error = store.writable('');
        const value = store.writable();
        const setError = ms > 0 ? debounce({ delay: ms }, error.set) : error.set;
        const setTouched = ms > 0 ? debounce({ delay: ms }, touched.set) : touched.set;
        const setDirty = ms > 0 ? debounce({ delay: ms }, dirty.set) : dirty.set;
        const schema = parentSchema.shape[name];
        const handleError = (e) => setError(getErrorMessage(e));
        try {
            initialValue = schema.parse(initialValue);
        }
        catch (e) {
            handleError(e);
        }
        const resetValue = () => value.set(initialValue);
        resetValue();
        const setValue = ms > 0 ? debounce({ delay: ms }, value.set) : value.set;
        const updateValue = ms > 0 ? debounce({ delay: ms }, value.update) : value.update;
        const handleChange = (e) => {
            setError('');
            let nextVal = e;
            if (e instanceof Event)
                if (e instanceof CustomEvent)
                    if (typeof e.detail === 'object' && 'value' in e.detail)
                        nextVal = e.detail.value;
                    else
                        nextVal = e.detail;
                else if (e.target instanceof HTMLInputElement)
                    if (e.target.type === 'checkbox')
                        nextVal = e.target.checked;
                    else
                        nextVal = e.target.value;
                else if (e.target instanceof HTMLSelectElement)
                    nextVal = e.target.value;
                else if (e.target instanceof HTMLTextAreaElement)
                    nextVal = e.target.value;
            try {
                nextVal = schema.parse(nextVal);
            }
            catch (e) {
                handleError(e);
            }
            setValue(nextVal);
            setTouched(true);
            setDirty(true);
        };
        const handleUpdate = (updater) => {
            const safeUpdater = (v) => {
                setError('');
                const updatedVal = updater(v);
                if (schema)
                    try {
                        return schema.parse(updatedVal);
                    }
                    catch (e) {
                        handleError(e);
                    }
                return updatedVal;
            };
            updateValue(safeUpdater);
            setTouched(true);
            setDirty(true);
        };
        const handleBlur = () => setTouched(true);
        const handleReset = () => {
            touched.set(false);
            dirty.set(false);
            error.set('');
            resetValue();
        };
        this.name = name;
        this.value = toReadable(value);
        this.touched = toReadable(touched);
        this.dirty = toReadable(dirty);
        this.error = toReadable(error);
        this.valid = store.derived(error, (e) => !e);
        this.handleChange = handleChange;
        this.handleUpdate = handleUpdate;
        this.handleBlur = handleBlur;
        this.reset = handleReset;
        this.setError = setError;
        this.setTouched = setTouched;
    }
}

/**
 * Instance that hold all our form's state, as Svelte's Store
 */
class ZodFormStore {
    /** Form's data. Will be passed to onSubmit handler */
    model;
    /** Form settings. Should not be update */
    options;
    /** Generated fields's stores */
    fields;
    /**
     * Function to start parsing, validating and submit the form's data.
     *
     * Should assign to form's submit event or submit button.
     *
     * ```svelte
     * <form on:submit={form.triggerSubmit}>
     *  ...
     * </form>
     * ```
     */
    triggerSubmit;
    /**
     * Function to reset the form to original state.
     *
     * Should assign to form's `reset` event, reset button, or call directly.
     *
     * ```svelte
     * <form on:reset={form.reset}>
     *  ...
     * </form>
     * ```
     */
    reset;
    /** Indicate if the form is being submitted (`onSubmit` handler is resolving).  */
    submitting;
    /**
     * Array of string contains all error messages
     * (including fields's errors and error return from `onSubmit` handler).
     */
    errors;
    /** Error message returned from `onSubmit` handler, or custom validation message. */
    error;
    /** Indicate if the form is edited or submitted. */
    dirty;
    /** Indicate if the form is valid. */
    valid;
    constructor(
    /** Zod's schema for data in the form. Should be a `ZodObject`. */
    schema, 
    /** Additional configuration */
    options) {
        this.options = options || {};
        // Form stores
        const submitting = store.writable(false);
        const dirty = store.writable(false);
        const formError = store.writable();
        // Fields stores
        const fieldNames = Object.keys(schema instanceof zod.ZodEffects ? schema.innerType().shape : schema.shape);
        const fieldStores = fieldNames.map((name) => new ZodFieldStore(schema instanceof zod.ZodEffects ? schema.innerType() : schema, name, this.options.initialValue?.[name], this.options.debounceDelay));
        const model = store.derived(fieldStores.map((z) => z.value), (z) => Object.fromEntries(zip(fieldNames, z)));
        const fieldErrors = store.derived(fieldStores.map((z) => z.error), (e) => e.filter(trurthly));
        const triggerSubmit = async () => {
            dirty.set(true);
            fieldStores.forEach((f) => f.setTouched(true));
            if (!this.options.onSubmit)
                return;
            const _fieldErrors = store.get(fieldErrors);
            if (_fieldErrors.length > 0) {
                if (this.options.debug)
                    console.debug('❗️ Form not valid', _fieldErrors);
                return;
            }
            formError.set('');
            submitting.set(true);
            try {
                const values = schema.parse(store.get(model));
                const result = await this.options.onSubmit(values);
                if (result)
                    formError.set(result);
            }
            catch (e) {
                if (this.options.debug)
                    console.debug(`🚫 Form submit error`, e);
                // Set Error back to field
                if (e instanceof zod.ZodError)
                    for (const issue of e.issues)
                        this.setError(issue.message, issue.path);
                formError.set(getErrorMessage(e));
            }
            finally {
                submitting.set(false);
            }
        };
        function handleReset() {
            dirty.set(false);
            submitting.set(false);
            formError.set('');
            fieldStores.forEach((f) => f.reset());
        }
        if (this.options.autoSubmitAfter && this.options.autoSubmitAfter > 0) {
            const debouncedTriggerSubmit = debounce({ delay: this.options.autoSubmitAfter }, () => {
                if (this.options.debug)
                    console.debug('⏱️ Auto submitting...');
                triggerSubmit();
            });
            let prevModel = null;
            model.subscribe((nextModel) => {
                if (prevModel === null) {
                    prevModel = nextModel;
                    return;
                }
                if (prevModel === nextModel)
                    return;
                prevModel = nextModel;
                debouncedTriggerSubmit();
            });
        }
        this.fields = Object.fromEntries(fieldStores.map((z) => [z.name, z]));
        this.triggerSubmit = triggerSubmit;
        this.reset = handleReset;
        this.submitting = toReadable(submitting);
        this.error = toReadable(formError);
        this.dirty = store.derived([dirty, ...fieldStores.map((a) => a.dirty)], (a) => a.some((b) => b));
        this.errors = store.derived([fieldErrors, formError], (errors) => errors.flatMap((z) => z).filter(trurthly));
        this.model = model;
        this.valid = store.derived(this.errors, (e) => !e.length);
    }
    /** Set the form's error message manually */
    setError(errorMessage, path) {
        const [currentPath] = path;
        const field = this.fields[currentPath];
        if (!field)
            return;
        field.setError(errorMessage);
    }
}

exports.ZodFormStore = ZodFormStore;
//# sourceMappingURL=index.cjs.map
